import time
import SEP.util
import re

class sep_machine:
  """class containing list of potential machines to run on"""



  def __init__(self,machfile="mach",njobs=9999,verb=None):
    self.verb=verb
    self.machines=self.init_list_file(machfile)
  
  def init_list_file(self,machinefile):
    """Get the list of machines from a file"""
    try: f = open (machinefile)
    except: raise SEP.util.err,"Trouble opening file:"+file
    lines=f.readlines()
    nocommentS=re.compile('^(.+)\#')    #remove commented material
    machnS=re.compile('^(\S+):(\S+)')    #remove commented material
    machS=re.compile('^(\S+)')    #remove commented material
    self.mach_list=[]
    self.status={}
    self.machname={}
    for line in lines:
       #remove the commented material
       nocomment=nocommentS.search(line.strip())
       if nocomment: do=nocomment[0]
       else: do=line.strip()
       ncount=0;
       machn=machnS.search(do)
       mach=machS.search(do)
       if machn:
         ncount=machn.group(2)
         mach=machn.group(1)
       if mach:
         ncount=1
         mach=mach.group(1)
       for i in range(ncount):
         name=mach+"-"+str(i)
         self.mach_list.append(name)
         self.status[name]="free"
         self.machname[name]=mach
    f.close()
    self.nmach=len(self.mach_list)

  def free(self):         
    """Return the list of free nodes"""
    mach={}
    for key,val in self.status.items():
      if val == "free": 
        mach[key]=self.machname[key]
    return mach

  def assign(self,mach_label):         
    """Mark that a node is being issued a job"""
    for mach in mach_label.values():
      self.status[mach]="sent:"+mach+":first"

  def record(self,jobid,machid,extra):         
    """Record that a job has effetively started on a node"""
    self.status[machid]="running:"+jobid+":"+extra


  def run_list(self):         
    """Return the list of nodes that have jobs running and submitted but not running"""
    sent_list=[]
    running_list=[]
    for key,val in self.status.items():
      stat=val.split(':')
      if stat[0] == "sent":
        if stat[2] == "notregistered": sent_list.append(self.machname[key])
        else: self.status[key]="sent:"+stat[1]+":notregistered"
      if stat[0] == "running": running_list.append(self.machname[key])
    sent_list=SEP.util.condense(sent_list) 
    running_list=SEP.util.condense(running_list) 
    return sent_list,running_list
  
    
  def dead(self,machid):
    """Mark node as dead""" 
    mach=self.machnames[machid]
    if self.verb: print "Marking %s as dead"%mach
    for key, val in self.machnames.items():
      if val == mach: self.status[key]="dead"

  def finish(self,mach_list):         
    """Mark that a job is finished"""
    for mach in mach_list: self.status[mach]="free"


