import SEP.sep_socket
import SEP.superset
import SEP.datapath
import SEP.mpi
import os
import sys
import commands
import SEP.util
 
class parfile:
  """ Define a sep parallel file"""
  def __init__(self,
     name=None,  #file name
     tag=None,  #tag name
     axis=None,  #axis
     isect_dict=None, #dictionary section ident:isect
     usage=None,  #usage INPUT,OUTPUT
     type="BLOCK", #distribution type BLOCK,SPREAD,COPY
     combine="YES",#combine the dataset when finished
     par_file=None, #parfile if it  already exists
     restart="NO",  #whether or not we are restarting a job
     shared="NO",  #whether or not the COPY distributed file are shared on a node
     verb=None     #whether or not to be verbose
     ):
 
     self.set_init_pars(name,tag,axis,type,usage,combine,par_file,restart,shared,isect_dict,verb)
     dead=0
     self.check_param_logic(dead)
     self.check_file(dead)
     if dead ==1: raise SEP.util.err," "
     if self.verb: print "Finished initializing ",self.name




  def tags(self,mach_list):
    """Get local section tag given machine  list section list"""
    sect_names=self.get_section_names(mach_list)
    tag_list={}
    for key in mach_list.keys():
      self.sect_dict[key]=sect_names[key]
      self.mach_dict[key]=mach_list[key]
      tag_list[key]=self.tag+sect_names[key]
    return tag_list

  def finish(self):
    """Finish processing a file"""
    nout=0
    if self.usage == "INPUT": return
    for n in self.naxis: nout=int(n)+int(nout)
    n,o,d,label,unit=self.par_file.axis(self.axis)
    self.par_file.set_axis(self.axis,nout,self.oaxis,d,label,unit)
    if self.type!="COPY": 
      self.file.set_axis(self.axis,nout,self.oaxis,d,label,unit)
    self.par_file.write_file()
    if self.combine == "YES": self.file.write_file()
     


  def done(self,key_list):
    """Mark that a section has been processed"""

    if self.usage == "INPUT": return
    for key in  key_list:
       self.mark_done(key)
    self.par_file.write_file()
    if self.combine == "YES": self.file.write_file()

  def dead(self,mach):
    """Mark that a node has died"""
    for key,val in self.mach_dict:
      if val == mach:
        isect=self.isect_dict[key]
        self.naxis[isect]=0
        self.par_file.set_fake(isect,1)

  def mark_done(self,key):
    """Mark section finished and update files"""
    sect=self.sect_dict[key]
    mach=self.mach_dict[key]
    isect=self.isect_dict[key]
    temp_file=SEP.sepfile.sep_file(remote=mach+":"+sect)
    self.par_file.add_dff_par("mach_sect"+str(isect),mach)
    self.par_file.add_dff_par("tag_sect"+str(isect),sect)
    self.par_file.set_fake(isect,0)
    ndim=temp_file.ndims()
    n,o,d,label,unit=temp_file.axis(self.axis)
    self.naxis[isect]=n
    if isect == 0: self.oaxis=o
    for i in range(ndim):
      iax=i+1
      n,o,d,label,unit=temp_file.axis(iax)
      if iax == self.axis and self.type != "COPY":
        if self.type == "BLOCK":
          o=float(o)-float(d)*int(isect)*int(n)
          n=int(n)*int(self.nsect)
        if self.type == "SPREAD":
          o=float(o)-float(d)*int(isect)
          d=float(d)/float(self.nsect)
          n=int(n)*int(self.nsect)
      self.par_file.set_axis(iax,n,o,d,label,unit)
      if self.combine == "YES": self.file.set_axis(iax,n,o,d,label,unit)
    if self.type == "COPY":
      self.par_file.set_axis(self.axis,self.nsect,0.,1.,"none","none")
     
  def get_section_names(self,mach_list):
    """Get local section names given machine list section list"""
    mach_dist_list={}
    sect_dist_list={}
    isect_dist_list={}
    sect_list={}
    for key,mach in mach_list.items():
      sect, dist=self.get_name(self.isect_dict[key],mach)
      if dist:  #if we need to distribute it
        mach_dist_list[key]=mach_list[key]
        sect_dist_list[key]=sect
        isect_dist_list[key]=self.isect_dict[key]
      sect_list[key]=sect
      self.sect_dict[key]=sect
      self.mach_dict[key]=mach_list[key]
    if len(mach_dist_list) >0:
      self.par_file.write_file()
      if self.verb: print "Distributing  tag=%s"%self.name,mach_dist_list
      self.mpi.distribute(self.name+"_par",self.name,
        isect_dist_list.values(),sect_dist_list.values(),mach_dist_list.values(),self.axis)
    return sect_list

  def get_name(self,isect,mach):
    """Get the section name given the machine and section number"""
    create=None
    if self.type=="COPY" and  self.shared=="YES":
      ilist,list=self.par_file.isect_tag_by_host(mach)
      if ilist == None: create=1
      else: sect_out=list[0]
    else: 
      mach_out,sect_out=self.par_file.mach_tag_by_isect(isect)
      if mach_out==None: create=1
      elif mach!=mach_out:  create=1
    if create:
      if  self.usage=="INPUT": sect_out=self.par_file.create_section(mach,isect)
      else:
        create=None
        sect_out=self.par_file.sect_name(mach,isect)
    return  sect_out,create
 
      
  def collect(self,key_list,extra_args):
    """Combine the sections in isect_list to output image"""
    isect_list=[]
    mach_list=[]
    if self.usage == "INPUT": return
    for key in key_list: isect_list.append(self.isect_dict[key])
    for key in key_list: mach_list.append(self.mach_dict[key])
    if self.verb: print "tag=%s combining sections"%self.name,key_list
    if self.combine=="YES":
      self.mpi.collect(self.par_file.name,self.file.name,
        isect_list,mach_list,self.axis,extra_args)

##########################
#INITIALIZATION FUNCTIONS#
##########################
                                                                                       
  def set_init_pars(self,name,tag,axis,type,usage,combine,par_file,restart,shared,isect_dict,verb):
     """ Initialize structure parameters"""
     self.par_file=None #parallel sepfile
     self.file=None     #non-parallel sepfile
     self.name=name
     self.tag=tag
     self.verb=verb
     self.axis=axis
     self.type=type
     self.usage=usage
     self.combine=combine
     self.isect_dict=isect_dict
     if par_file != None: self.par_file=par_file
     self.mpi=SEP.mpi.mpi(verb=verb)
     self.shared=shared
     self.restart=restart
     self.sect_dict={}
     self.mach_dict={}
     self.naxis=[]
     self.oaxis=0.
                                                                                       
  def check_param_logic(self,dead):
     """Check to make sure required parameters are set"""
     if self.name == None: raise SEP.util.err, "Must supply name"
     if self.isect_dict == None: raise SEP.util.err, "Must supply isect_dict_list"
     self.nsect=len(self.isect_dict)
     if self.tag == None: raise SEP.util.err, "Must supply tag"
     if self.axis == None: raise SEP.util.err, "Must supply axis"
     if self.type != "BLOCK" and self.type!="SPREAD" and self.type!="COPY":
       raise SEP.util.err, "Distribution type must be BLOCK,SPREAD, or COPY type=",type
     if self.usage !="INPUT" and self.usage!="OUTPUT":
       raise SEP.util.err, "Distribution type must be INPUT or OUTPUT  usage="+self.usage
     for i in range(len(self.isect_dict.keys())): self.naxis.append(0)
                                                                                       
  def check_file(self,dead):
     """Check to make sure provided files are logical"""
     if self.usage == "INPUT" or self.restart=="YES":
       if not os.path.isfile(self.name):
         raise SEP.util.err, "Input file["+self.name+"] doesn't exist"
       else: self.file=SEP.superset.superset(file=self.name)
     else:
       self.file=SEP.superset.superset(name=self.name)
       for i in range(self.axis-1):
         self.file.set_axis(i+1,1,0.,1.,"none","none")
       self.file.set_axis(self.axis,self.nsect,0.,1.,"none","none")
       self.file.write_file()
       
     if self.restart=="NO":
       if self.file.dff != None:
         raise SEP.util.err,"Input shouldn't be a distributed file unless restarting"
       else:
         self.par_file=SEP.superset.superset(supersetfile=self.file,
           name=self.name+"_par")
         self.par_file.create_dff(self.axis,self.nsect,self.type)
         self.par_file.write_file()
     else:                                                 
       if not os.path.isfile(name+"_par"):
         raise SEP.util.err, "Parallel file["+name+"_par] doesn't exist"
       else: 
         self.file=SEP.superset(file=self.name+"_par")
         naxis=self.par_file.naxis_isect()

