import SEP.sep_socket
import SEP.machine
import os
import sys
import string

class parjob:


  def __init__(self, program=None,files=None,pars=None,
    status=None,timeinterv=300,ip="192.168.100.2",machfile="mach",max_jobs=9999,verb=None,command_func=None):
    """INitialize a parallel job"""
    if None==program and command_func==None: 
      SEP.util.err,"Must provide program or command_func"
    if None==files: SEP.util.err,"Must provide files"
    if None==pars: SEP.util.err,"Must provide par"
    self.mach=SEP.machine.sep_machine(machfile,max_jobs,verb)
    self.pars=pars
    self.program=program
    self.files=files
    self.verb=verb
    self.shell="rsh"
    self.s_ip="192.168.100.2"
    self.status=status
    self.command_func=command_func
    self.timeinterv=300
    self.mach_parts={}
    if self.status==None:
      self.setup_status()
    elif len(self.status)!= len(parts):
      raise SEP.util.err,"Length of status and parts must be equiv"
    self.check_files()

  def check_files(self):
    """ Check to make sure parallel files make sense"""
    found=0
    nsect=len(self.pars)
    for file in self.files:
      if file.type != "COPY":
        if nsect!=file.nsect:
          err= "Number of sections  %d not equal to number of parts %d in file=%s" % (nsect,file.nsect,file.name)
          raise SEP.util.err,err
        if found==0:
          type=file.type
          nsect=file.nsect
          found=1
        else:
          if type!=file.type:
            raise SEP.util.err, "File types not the same for all paralllel files"

  def setup_status(self):
    """ Mark the status on all parts to todo """
    self.status={}
    for part in self.pars.keys(): self.status[part]="todo:none"

  def get_to_do(self,machs):
    """ return the sections we want to process"""
    ifound=0;
    nmach=len(machs)
    nlen=len(self.status)
    machlabel=machs.keys()
    machname=machs.values()
    mach_out={}
    mach_label={}
    parts={}
    key_out=[]
    if nlen==0: 
      raise SEP.util.err,"Eventually wait until a machine is ready"
      
    for key in self.pars.keys():
      if self.status[key] == "todo:none":
        parts[key]=self.pars[key]
        self.mach_parts[key]=machlabel[ifound]
        mach_out[key]=machname[ifound]
        mach_label[key]=machlabel[ifound]
        self.status[key]="sent:"+machlabel[ifound]
        ifound=ifound+1
        if ifound == nmach: return parts,mach_out,mach_label
    return parts,mach_out,mach_label
        
 

  def commands(self,mach):
    """Build the remote commands to run for the parallel job"""

    run_this=[]
    parts,machs,labels=self.get_to_do(mach)



    tag_job={}
    for key in parts.keys():
      tag_job[key]=" "

    for file in self.files:
      tags=file.tags(machs)
      for key,tag in  tags.items():
        tag_job[key]=tag_job[key]+" "+tag

    server_string=" sep.master_port=%d sep.master_ip=%s "%(self.s_num,self.s_ip)
    for key in parts.keys():
      job_string="%s sep.jobid=%s sep.mach_label=%s"%(server_string,key,labels[key])
      if self.verb:print "SENDING JOB %s to %s pars=%s"%(key,machs[key],parts[key])
      if self.command_func:
        run=self.command_func(key,job_string,tag_job[key].split())
      else:
        run="%s %s sep.begin_par=1 sep.end_par=1  %s %s"%(self.program,job_string,self.pars[key],tag_job[key])
      run=self.fix_rsh_com(machs[key],run)

      run_this.append(run)
      if self.verb>1: print "preparing to run:\n\n",run
    return run_this,labels

  def fix_rsh_com(self,mach,inp):
    list=inp.split()
    for c in list: 
      if c == '"': c="'"
    out=self.shell+" "+mach+' "'+string.join(list)+'"'
    return out
  
    
  

  def  run_job(self):
    """ Start job loop """

    #get info on the server
    self.socket=SEP.sep_socket.sep_server(self.action,self.mach,self.timeinterv)

    self.s_num=self.socket.socket_number()


#    if not  self.start_jobs():
#      raise SEP.util.err, "Trouble starting initial jobs"

  
    commands=self.build_commands()
    if commands== None: raise SEP.util.err,"Can't build initial commands"
    self.socket.run(commands)



  def record(self,jobid,machid,extra):
     """" Record that a task is running"""
     self.mach.record(jobid,machid,extra) #record that we started a job
     self.status[jobid]="running:"+machid+":"+extra
     return 1
      

  def start_jobs(self):
     """ Start a series of jobs """
     commands=self.build_commands()
     if commands != None and len(commands)>0: 
       SEP.util.execute(commands)
       return 1
     else: return None
        
         


  def build_commands(self):
     """Build commands to run on free nodes"""
     machines=self.mach.free()  #get initial list of machines
     commands,mach_label=self.commands(machines)
     if commands == None: return None
     self.mach.assign(mach_label)  #assign machines
     return commands

      
  def finish(self,jobid,machid,extra):
     """ Mark job as finished"""
     self.mach.finish([machid])      #machine is free
     for file in self.files: file.done([jobid])
     self.status[jobid]="finished:"+machid
     if not self.start_jobs():
       remain=None
       for val in self.status.values():
         stat,id=val.split(":")
         if stat != "finished" and stat!="collected": remain=1
       if not remain: 
         for file in self.files: file.finish()
         self.collect()
       return remain
     return 1  
     

  
  def collect(self):
     """Collect the data for all nodes where a process isn't currently running"""
     sent,running=self.mach.run_list()
     i=0;
     key_list=[]
     for key,val in self.status.items():
       stat,mach=val.split(":")
       if running.count(mach) ==0 : key_list.append(key)
     if len(key_list)>0:
       for file in self.files: file.collect(key_list,[" "])

     
  def dead(self,machdead):
     """Mark a node as dead"""
     for key,val in self.mach_parts.items():
       stat,mach=self.status[key].split(":")
       if machdead == mach and stat != "collected":
        status[key]="todo:None"
     return 1
       
      
  def action(self,msg):
     """Perform actions based on node returns"""
     oper,jobid,machid,extra=self.decode_action(msg)
     if   oper== 'record':  status=self.record(jobid,machid,extra)
     elif oper== 'finish':  status=self.finish(jobid,machid,extra)
     elif oper== 'dead'  :  status=self.dead(machid)
     return status


  def decode_action(self,msg):
     """Decode the actions returned by the server"""
     parts=msg.split(":")
     return parts[0],parts[1],parts[2],parts[3]
     
