import re
import string
import pwd
import sys
import os
import socket
import time
import SEP.par
import commands
from math import *

class format_file(SEP.par.sep_pars):
  """Extension of sep_par class that also knows about axes, in"""

  def __init__(self,file=None,name=None,sepfile=None):
    SEP.par.sep_pars.__init__(self,file,name,sepfile)

  def axis(self,iax,ext=None):
     """Return axis in a file"""
     if ext == None: e=""
     else:  e=ext
     n=1
     o=0
     d=1
     unit="none"
     label="none"
     val,err = self.par("n"+str(iax)+e)
     if err == 1 : n= val
     val,err = self.par("o"+str(iax)+e)
     if err == 1 : o= val
     val,err = self.par("d"+str(iax)+e)
     if err == 1 : d= val
     val,err = self.par("label"+str(iax)+e)
     if err == 1 : label= val
     val,err = self.par("unit"+str(iax)+e)
     if err == 1 : unit= val
     return n,o,d,label,unit

  def set_axis(self,iax,n,o,d,label,unit,e=None):
    """Set axis in a file"""
    if e == None: suf=""
    else : suf=e
    self.add_param('n'+str(iax)+suf,n,type="int")
    self.add_param('o'+str(iax)+suf,o)
    self.add_param('d'+str(iax)+suf,d)
    self.add_param('label'+str(iax)+suf,label)
    self.add_param('unit'+str(iax)+suf,unit)

  def ndims(self,suf=None,ibeg=1):
    """Number of dimensions in file"""
    if suf == None: e=""
    else : e=suf
    if ibeg==None: i=ibeg
    else: i=1
    iax=i; err=1; not1=0
    while err ==1:
       s="n"+str(iax)+e
       val,err = self.par(s)
       if err == 1 and int(val) !=1: not1=iax
       iax=iax+1
    return not1  

  def create_fake_in(self,name=None,file=None,host=None):
    """Create a fake data file """
    if name==None: nm=SEP.datapath.datafile(self.name,host)
    else: nm=name

    if file==None: fl=self.name
    else: fl=file

    self.add_param("in",nm)
    try: f = open (fl,"w")   #OPEN THE FILE
    except: raise SEP.util.err, 'Trouble opening file '+ fl
    f.write(" ") 
    f.close()

class history(format_file):
  """Extension of the format_file class """
  def __init__(self,file=None,name=None,sepfile=None):
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)

class hff(format_file):
  """Extension of the format_file class """
  def __init__(self,file=None,name=None,sepfile=None):
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)
    self.nkeys=0

  def set_nkey(self,nkey):
    self.nkeys=nkey
    self.add_param("n1",nkey,"int")

  def key(self,ikey):
     """Return key from a hff file"""
     list=("hdrkey","hdrtype","hdrfmt")
     out=[]
     for name in list:
       grab=name+str(ikey)
       val,err = self.par(grab)
       if err==1: SEP.util.err, grab+" doesn't exist in file"
       out.append(val)
     return out[0],out[1],out[2]

  def set_key(self,name,type,fmt,ik=None):
     """Set key in a hff file"""
     if ik==None: ikey=self.nkeys+1
     else: ikey=ik
     if ikey > self.nkeys: self.set_nkey(ikey)
     if ikey== None: SEP.util.err,"ikey must be specified" 
     if name== None: SEP.util.err,"name must be specified" 
     if type== None: SEP.util.err,"type must be specified" 
     if fmt== None: SEP.util.err,"fmt must be specified" 
     self.add_param("hdrkey"+str(ikey),name)
     self.add_param("hdrtype"+str(ikey),type)
     self.add_param("hdrmt"+str(ikey),fmt)


class gff(format_file):
  """Extension of the format_file class """
  def __init__(self,file=None,name=None,sepfile=None):
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)

  def axis(self,iax,ext):
    if iax ==1 : SEP.util.err,"Can't request first axis from a grid"
    SEP.sepfile.format_file.axis(iax,"_grid")

  def set_axis(self,iax,ext,n,o,d,label,unit):
    if iax ==1 : SEP.util.err,"Can't set first axis from a grid"
    SEP.sepfile.format_file.set_axis(iax,n,o,d,label,unit,"_grid")
 
  def ndims(self):
    SEP.sepfile.ndims.axis("_grid",2)




class sep_file:

  def __init__(self,file=None,sepfile=None,name=None,fmt=None,remote=None):
    self.grid=None
    self.header=None
    if remote!=None: self.remote_init(remote)
    elif file != None: self.file_init(file) 
    elif sepfile != None: 
      if name !=  None: self.sep_init(sepfile,name)
      else:
        raise SEP.util.err, "Must provide the name if initializing from another sepfile"
    elif name !=None: self.name_init(name,fmt)
    else:
      raise SEP.util.err, "Must initialize a sep_file with a file or another sep_file"
    self.name=self.history.name


  def remote_init(self,remote):
    """Initialize a file from a remote location"""
    pts=remote.split(":")
    if len(pts)!=2: SEP.util.err, "Expecting remote location of the form host:file"
    mach=pts[0]
    file=pts[1]
    temp_file=SEP.util.temp_file("temp_rep")
    SEP.util.execute(["rcp "+mach+":"+file+" "+temp_file])
    self.history=SEP.sepfile.history(temp_file)
    list={'hff':"@@",'gff':"@@@@"}
    cp_com=[]
    rm_com=["rm "+temp_file]
    for part,ext in list.items():
      val,err=self.history.par(part)       
      if err ==1: 
        rm_com.append("rm "+temp_file+ext)
        cp_com.append("rcp "+mach+":"+val+" "+temp_file+ext)
    SEP.util.execute(cp_com)
    val,err=self.history_par("hff")
    if err==1 and val != "-1":
      self.hff=SEP.sepfile.hff(temp_file+"@@")
      val,err=self.history_par("gff")
      if err==1 and val != "-1":
        self.gff=SEP.sepfile.gff(temp_file+"@@@@")
        self.file_format="GRID"
      else:  self.file_format="HEADERS"
    else: self.file_format="REGULAR"
    SEP.util.execute(rm_com)

  def name_init(self,name,fmt):
    """Initialize a blank file"""
    if fmt==None: self.file_format="REGULAR"
    else: self.file_format=fmt
    if self.file_format != "REGULAR" and  self.file_format != "HEADERS" and  self.file_format != "GRID":
     raise SEP.util.err, fmt+"= format must be REGULAR,HEADERS,or GRID"
    self.history=SEP.sepfile.history(name=name)
    self.history.set_axis(1,1,0.,1.,"none","none")
    if self.file_format=="REGULAR": 
      self.history.set_axis(2,1,0.,1.,"none","none")
    else:
      self.headers=SEP.sepfile.hff(name=name+"@@")
      self.headers.set_axis(2,1,0.,1.,"none","none")
      self.headers.set_key("none","scalar_float","xdr_float")
      if self.file_format=="GRID":
        self.grid=SEP.sepfile.gff(name=name+"@@@@")
        self.grid.set_axis(2,1,0.,1.,"none","none")

  def sep_init(self,sepfile,name):
    self.history=SEP.sepfile.history(sepfile=sepfile.history,name=name)
    self.file_format=sepfile.file_format
    if sepfile.file_format != "REGULAR":
      self.headers=SEP.sepfile.hff(seppar=sepfile.headers,name=name+"@@")
      if sepfile.file_format == "GRID":
        self.grid=SEP.sepfile.gff(seppar=sepfile.grid,name=name+"@@@@")

  def file_init(self,file):
    try: f = open (file)   #OPEN THE FILE
    except: raise SEP.util.err, 'Trouble opening file:'+ file
    f.close()
    self.history=SEP.sepfile.history(file=file)
    val, err = self.history.par("hff")
    if err == 1 and val != "-1":
      self.header=SEP.sepfile.hff(file=val)
      val, err = self.history.par("gff")
      if err ==1 and val != "-1":
        self.grid=SEP.sepfile.grid(file=val)
        self.file_format="GRID"
      else: self.file_format="HEADERS"
    else: self.file_format="REGULAR"

  def history_par(self,par):
     """Get parameter from history file"""
     return self.history.par(par)

  def header_par(self,par):
     """Get parameter from hff file"""
     if self.file_format == "REGULAR": return None,-1
     return self.header.par(par)

  def grid_par(self,par):
     """Get parameter from gff file"""
     if self.file_format != "GRID": return None,-1
     return self.grid.par(par)

  def add_history_par(self,par,value,type="str"):
     """Set parameter in history file"""
     return self.history.add_param(par,value,type)

  def add_header_par(self,par,value,type="str"):
     """Set parameter in hff file"""
     if self.file_format == "REGULAR": return None,-1
     return self.header.add_param(par,value,type)

  def add_grid_par(self,par,value,type="str"):
     """Set parameter in gff file"""
     if self.file_format != "GRID": return None,-1
     return self.grid.add_param(par,value,type)

  def axis(self,iax):
     """Set axis for SEP dataset"""
     if self.file_format == "GRID" and iax != 1:  
       return self.grid.axis(iax)
     elif self.file_format == "HEADERS" and iax !=1:  
       return self.header.axis(iax)
     else :  
       return self.history.axis(iax)

  def write_file(self,base=None):
    """Write a seplib file to disk"""
    if self.file_format!="REGULAR":
      if base == None: self.header.write_pars(self.header.name)
      else:
        nmh,val=self.history.par("hff")
        self.history.add_param("hff",base+"@@")
        self.header.write_pars(base+"@@")
        if self.file_format=="GRID":
          if base == None: self.grid.write_pars(self.header.name)
          else:
            nmg,val=self.history.par("gff")
            self.history.add_param("gff",base+"@@@@")
            self.grid.write_pars(base+"@@@@")
    if base == None: self.history.write_pars( self.history.name)
    else: self.history.write_pars(base)
    if self.file_format  != "REGULAR" and base != None:
      self.history.add_param("hff",nmh)
      if self.file_format=="GRID" : self.grid.add_param("gff",nmg)

  def set_axis(self,iax,n,o,d,label,unit):
     """Set an axis in a seplib dataset"""
     if self.file_format == "GRID":  
       return self.grid.set_axis(iax,n,o,d,label,unit)
     elif self.file_format == "HEADERS":  
       return self.header.set_axis(iax,n,o,d,label,unit)
     else :  return self.history.set_axis(iax,n,o,d,label,unit)
     
  def create_fake(self,host=None):
    """Open up 0 length files and set in="""
    self.history.create_fake_in(host=host)
    if self.header != None: self.header.create_fake_in(host=host)
    if self.grid != None: self.grid.create_fake_in(host=host)

  def remove(self):
    """Remove the format and binary files from disk if they exist"""
    parts=[self.history,self.header,self.grid]
    list=[]
    for part in parts:
      if part != None:
        if os.path.isfile(part.name): list.append("rm "+part.name)
        val,err = part.par("in")
        if err==1: 
          if os.path.isfile(val): list.append("rm "+val)
    SEP.util.execute(list)
     

  def send(self,host,file,binary):
     """Transfer a sepfile to another host"""
     temp=SEP.util.temp_file("temp_rep")
     tempstruct=SEP.sepfile.sep_file(sepfile=self,name=temp)
     copy_list={}
     copy_list[tempstruct.history.name]=file
     if binary=="YES":
         bin=SEP.datapath.datafile(file,host)
         val,err = self.history.par("in")
         if err == 1:  
           copy_list[val]=bin
           tempstruct.history.add_param("in",bin)
     erase_com=["rm "+temp]
     if tempstruct.header != None:
        erase_com.append("rm "+temp+"@@")
        copy_list[tempstruct.header.name]=file+"@@"
        if binary=="YES":
           bin=SEP.datapath.datafile(file+"@@",host)
           val,err = self.header.par("in")
           if err == 1:  
             copy_list[val]=bin
             tempstruct.header.add_param("in",bin)
        tempstruct.history.add_param("hff",file+"@@")
     if tempstruct.grid != None:
        erase_com.append("rm "+temp+"@@@@")
        copy_list[tempstruct.grid.name]=file+"@@@@"
        if binary=="YES":
           bin=SEP.datapath.datafile(file+"@@@@",host)
           val,err = self.grid.par("in")
           if err == 1:  
             copy_list[val]=bin
             tempstruct.grid.add_param("in",bin)
        tempstruct.history.add_param("gff",file+"@@@@")
     tempstruct.write_file(temp)
     copy_com=[]
     for myl,r in copy_list.items():
       copy_com.append("rcp "+myl+" "+host+":"+r)
     SEP.util.execute(copy_com)
     SEP.util.execute(erase_com)

  def ndims(self):
     """Return the number of dimensions in a SEPlib3d dataset"""
     if self.file_format == "GRID":  return self.grid.ndims()
     elif self.file_format == "HEADERS":  return self.header.ndims()
     elif self.file_format == "REGULAR":  return self.history.ndims()
     else: raise SEP.util.err,"File format not specified"

