import re
import string
import pwd
import sys
import os
import time
import SEP.par
import SEP.sepfile
import SEP.datapath
import SEP.util
from math import *

class sep_section(SEP.sepfile.sep_file):


  def __init__(self,isect=None,mach=None,file=None,name=None,sepfile=None,fmt=None,fake=0):
    if isect == None: raise SEP.util.err, "Must specify isect"
    if mach == None: raise SEP.util.err, "Must specify mach"
    SEP.sepfile.sep_file.__init__(self,sepfile=sepfile,file=file,name=name,fmt=fmt) 
    self.isect=isect
    self.mach=mach
    self.fake=fake

class sep_dff(SEP.par.sep_pars):
  """Extension of sep_par class that handles the dff file"""
 
  def __init__(self,file=None,name=None,sepfile=None,nsect=None,pattern=None,axis=None):
    SEP.par.sep_pars.__init__(self,file,name,sepfile)
    if nsect==None:
      nsect,err=self.par("nsect")
      if err==0: nsect=0
    self.set_nsect(nsect)
    if pattern==None:
      pattern,err=self.par("pattern")
      if err==0: pattern="BLOCK"
    self.set_pattern(pattern)
    if axis==None:
      axis,err=self.par("axis")
      if err==0: axis=-1
    self.set_axis(axis)

  def set_naxis(self,n):
    """Set naxis value"""
    self.add_param("naxis_convert",n,"int")
   

  def set_section(self,isect,tag,mach,fake=0):
    self.add_param("tag_sect"+str(isect),tag)
    self.add_param("mach_sect"+str(isect),mach)
    if isect>= self.nsect:  self.set_nsect(isect+1)
    self.add_param("fake_sect"+str(isect),fake)

  def set_nsect(self,nsect):
    self.add_param("nsect",nsect,type="int")
    self.nsect=nsect

  def set_pattern(self,pattern):
    self.add_param("pattern",pattern)
    self.pattern=pattern

  def set_axis(self,axis):
    self.add_param("axis",axis,type="int")
    self.axis=axis


class superset(SEP.sepfile.sep_file):
  def __init__(self,file=None,supersetfile=None,name=None,sepfile=None,fmt=None):

    self.dff=None
    self.sections={}
    if supersetfile==None:
      SEP.sepfile.sep_file.__init__(self,file,sepfile,name,fmt)
      if  file!=None:
        dffname,err=self.history.par("dff")
        if err == 1 and dffname != "-1":
          SEP.dff=SEP.superset.sep_dff(name=name,file=dffname)
    else:
      SEP.sepfile.sep_file.__init__(self,file,supersetfile,name,fmt)
      if supersetfile.dff != None:
        SEP.dff=SEP.superset.sep_dff(name=name,sepfile=supersetfile.dff)

  def dff_par(self,par):
     """Grab a parameter from a dff file"""
     return self.dff.par(par)

  def add_dff_par(self,par,value,type="str"):
     """Set a parameter in a dff file"""
     return self.dff.add_param(par,value,type)

  def read_section(self,isect): 
     """Read a section from a file"""
     sect,err=self.dff_par("tag_sect"+str(isect))
     if err==0: raise SEP.util.err, str(isect)+"=isect sect_tag doesn't exist"
     mach,err=self.dff_par("mach_sect"+str(isect))
     if err==0: raise SEP.util.err, str(isect)+"=isect mach_sect doesn't exist"
     self.sections[str(isect)]=SEP.superset.sep_section(isect,mach,name=sect,file=sect)

  def set_fake(self,isect,fake):
     self.add_dff_par("fake"+str(isect),fake,type="int")

  def write_section(self,isect): 
     """Write a section to a file"""
     if self.sections.has_key(str(isect)):
       sect=self.sections[str(isect)]
       sect.write_file()
     else: raise SEP.util.err,str(isect)+"=section doesn't exist"

  def write_file(self,base=None):
    """Write a superset file to disk"""
    if base==None: file=self.name
    else: file=base
    SEP.sepfile.sep_file.write_file(self,base)
    if self.dff != None:
      self.dff.write_pars(file+"@@@@@@")
   

  def create_dff(self,axis,nsect,dtype="BLOCK"):
    """Create a dff file for the given dataset"""
    
    dffname=self.history.name+"@@@@@@"
    self.modify_sepfile(axis,nsect,dtype,dffname)
    self.dff=SEP.superset.sep_dff(name=dffname,
      nsect=nsect,pattern=dtype,axis=axis)
    self.dff.set_naxis(self.axis(axis)[0])

  def modify_sepfile(self,axis,nsect,dtype,dff_name):
    """Modify the sepfile according to dff parameters (and error checking)"""
    ndim=self.ndims()
    if axis < 1: raise SEP.util.err,"Axis must be greater than 0"
    if dtype=="COPY":
      if ndim >= axis:raise SEP.util.err,"Axis %d must be less than ndims=%d"%(axis,ndim)
      for i in range(axis-ndim-1):
        self.set_axis(i+ndim+1,1,0.,1.,"none","none")
      self.set_axis(axis,nsect,0.,1.,"none","none")
    elif dtype=="BLOCK" or dtype=="SPREAD":
      if ndim < axis: raise SEP.util.err,"Axis %d larger than ndims=%d"%(axis,ndim)
      n,o,d,label,unit=self.axis(axis)
      if n < nsect: 
        raise SEP.util.err,"nsect=%d is greater than axis(%d) length=%d file=%s"%(nsect,axis,n,self.history.name)
    else: raise SEP.util.err, dtype+"=dtype, I understand only COPY, BLOCK, and SPREAD"
    self.add_history_par("dff",dff_name)

  def sect_axis(self,isect):
    """Return the n,o,d,label,unit for the distributed axis given section number"""
    n,o,d,label,unit=self.axis(self.dff.axis)
    if self.dff.pattern == "BLOCK":
      j=1
      nt=0
      nremain=int(n)
      f=0
      i=0
      for i in range(isect+1):
        nt=ceil(int(nremain)/(int(self.dff.nsect)-i))
        nremain-=int(nt);
        f=int(nt)+int(f);
      nout=nt;
      f-=nt;
    elif dtype == "SPREAD":
      j=self.dff.nsect
      f=isect
      nout=floor(n/self.dff.nsect)
      nremain=n - nout*self.dff.nsect
      if nremain!=0:
        if isect < nremain: nout+=1;
    elif dtype == "COPY":
      nout=1
      f=0
      j=1
    else:
      raise SEP.util.err, "Uknown distribution type "+dtype
    oout=float(o)+float(d)*int(f)
    dout=float(d)*int(j)
    return nout,oout,dout,label,unit

  def create_section(self,host,isect):
    """Create a section """
    if host==None: mach=os.uname()[1]
    else: mach=host
    if isect == None: i=self.dff.nsect
    else: i=isect
    if self.sections.has_key(str(i)): del self.sections[str(i)]
    tag_out=self.sect_name(mach,i)
    tag_temp= SEP.datapath.datapath(os.uname()[1])+"_DFF_"+str(i)+"_"+os.path.basename(self.name)
    temp=SEP.superset.sep_section(i,mach,name=tag_temp,sepfile=self)
    temp.add_history_par("dff",-1,"int")
    temp.create_fake()
    n,o,d,label,unit=self.sect_axis(isect)
    temp.set_axis(self.dff.axis,n,o,d,label,unit)
    temp.write_file()
    temp.send(host,tag_out,binary="YES")
    temp.remove()
    self.sections[str(isect)]=SEP.superset.sep_section(i,host,name=tag_out,sepfile=temp)
    self.dff.set_section(isect,tag_out,mach,1)
    return tag_out

  def sect_name(self,mach,i):
    tag_out= SEP.datapath.datapath(mach)+os.path.basename(self.name)+"_DFF_"+str(i)
    return tag_out

  def naxis_isect(self):
    """Get the amount along the shared axis possessed by each section"""
    naxis=[]
    if self.dff == None: SEP.util.err,"Requesting naxis in non-distributed dataset"
    for i in self.dff.nsect:
      val,err = self.dff_par("fake"+str(i))
      if err ==0 or val==0:
        if self.sections.has_key(str(i)):
           temp=self.sections[str(i)]
        else:
          mach,err1=self.dff_par("mach_sect"+str(i))
          tag,err2=self.dff_par("tag_sect"+str(i))
          if err1!=1 or err2!=1:
             raise SEP.util.err,"invalid dff file, missing tag_sect or mach_sect"
          temp=SEP.sepfile.sep_file(remote=mach+":"+tag)
        n,o,d,label,unit=temp.axis(self.dff.axis)
        naxis.append(n)
      else: naxis.append(0)

  def create_sections(self):
    """Create all of the sections in a dataset"""
    host=os.uname()[1]
    for i in range(self.dff.nsect):
      self.create_section(host,i)

  def set_axis(self,iax,n,o,d,label,unit):
    """Set axis to given n,o,d,label, and unit"""
    SEP.sepfile.sep_file.set_axis(self,iax,n,o,d,label,unit)
    if self.dff != None:
      if self.dff.axis == iax: self.dff.set_naxis(n)
    
  def mach_tag_by_isect(self,isect):
    """Return the machine and tag of a given section number"""
    if self.sections.has_key(str(isect)):
      return self.sections[str(isect)].mach,self.sections[str(isect)].history.name
    else: return None,None

  def isect_tag_by_host(self,host):
    isects=[]
    tags=[]
    for key,val in self.sections.items():
      if val.mach == host:
        isects.append(int(key))
        tags.append(val.history.name)
    if len(isect)> 0: return isects,tags
    else: return None,None
