/*
=head1 NAME

 Combine
  
=head1 SYNOPSIS

  Combine intag= outtag= axis=

=head1 PARAMTERS

=over 4

=item assign - integer*

 section number to assign from each thread (starting with thread 1) 
   
 When running in mpi mode len(assign)=nmpi-1

=item max_size - integer

 [100]  maximum memory in megabyte to use on node (max memory will then be 2x)


=item axis - integer

![2]  axis to distribute over

=item verb - logical

  [0]  Wether to be verbose


=head1 CATEGORY
                                                                                                        
<util/cube>
                                                                                                        


=cut

*/

#include<seplib.h>
#include<sep3dc.h>
#include<sepaux.h>
#include<superset.h>
#ifndef SEPNULL2
#define SEPNULL2 ((void *) 0)
#endif
#ifdef SEP_MPI
#include<mpi.h>
#endif

#define HEAD "/dev/null"
                                                                                                                     
#define MY_SEP_DOC \
 sep_add_doc_line("NAME");\
 sep_add_doc_line("     Combine");\
 sep_add_doc_line("");\
 sep_add_doc_line("SYNOPSIS");\
 sep_add_doc_line("      Combine intag= outtag= axis=");\
 sep_add_doc_line("");\
 sep_add_doc_line("PARAMTERS");\
 sep_add_doc_line("    assign - integer*");\
 sep_add_doc_line("         section number to assign from each thread (starting with thread 1) ");\
 sep_add_doc_line("");\
 sep_add_doc_line("         When running in mpi mode len(assign)=nmpi-1");\
 sep_add_doc_line("");\
 sep_add_doc_line("    max_size - integer");\
 sep_add_doc_line("         [100]  maximum memory in megabyte to use on node (max memory will then be 2x)");\
 sep_add_doc_line("");\
 sep_add_doc_line("    axis - integer");\
 sep_add_doc_line("        ![2] axis to distribute over");\
 sep_add_doc_line("");\
 sep_add_doc_line("    verb - logical");\
 sep_add_doc_line("          [0]  Wether to be verbose");\
 sep_add_doc_line("");\
 sep_add_doc_line("CATEGORY");\
 sep_add_doc_line("        <util/cube>");\
 sep_add_doc_line("");

int main(int argc, char **argv){
sep3d input,output;
float *buf;
int   nh,i3,ndo,verb,max_size,n,nhout;
int   sect,nblock,axis,master;
int *nwind,*fwind,*jwind,*nsect,*fsect,*jsect;
int *nin,*fin,*jin,*nout,*fout,*jout;
int *ncopy, *fcopy,*jcopy;
char   pattern[256];
char *buf_in,*buf_out;
float *float_out,*float_in;
int esize,ns[2]; 
int master_data,i;
int assign[9999],nassign;
int write_head,write_grid;
int isect,iassign;
int copy,impi,nmpi,nloop;



#ifdef SEP_MPI
  MPI_Init(&argc,&argv);
   initpar(argc,argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nmpi);
  MPI_Comm_rank(MPI_COMM_WORLD, &impi);
  if (impi==0) {
    if(0==getch("verb","d",&verb)) verb=0;
    mpi_sep_send_args(nmpi,10,verb);
  }
  else mpi_sep_receive_args();
  master_data=0;
  sect=nmpi;
  if(master_data==0) sect=sect-1;
#else
   initpar(argc,argv);
  master_data=1;
  sect=2;
  if(0==getch("verb","d",&verb)) verb=0;
#endif
  init_3d();
  MY_SEP_DOC
  doc(SOURCE);


/*get parameters*/
                                                                                                                     
if(0==getch("max_size","d",&max_size)) max_size=50;

                                                                                                                     

/*initialize input*/
sep3d_initialize(&input);
getch_add_string("intag.reassign=0");
if(0!= init_sep3d_tag("intag",&input,"INPUT"))
 seperr("trouble initializing input \n");

if(input.file_format[0]=='G') {write_head=1; write_grid=1;}
else if(input.file_format[0]=='H') {write_head=1; write_grid=0;}
else { write_head=0; write_grid=0;}



if(0!= init_sep3d_tag("outtag",&output,"SCRATCH"))
 seperr("trouble initializing input \n");

if(input.nsect < 2)
  seperr("data not sectioned \n");

if(0!=sep3dc_grab_sect_axis(&input,&axis))
  seperr("trouble grabbing sectioned axis \n");

if(axis <= output.ndims) copy=0;
else copy=1;

nassign=getch("assign","d",assign);
sect=input.nsect;
if(nassign!=n-1) {
   fprintf(stderr, "assign length=%d nmpi=%d \n",nassign,n);
   seperr("only works in MPI mode number of assigns= nmpi-1\n");
}
for(i=0; i < nassign; i++){
   if(0!=sep3dc_assign_sect(&input,i+1,assign[i]))
    seperr("trouble assigning section \n");
} 
/*sep3dc_write_description("outtag",&output);*/



if(input.data_format[0]=='C') esize=8;
else if(input.data_format[0]=='B') esize=1;
else esize=4;

max_size=max_size*1000*1000/esize;
init_loop_calc(output.ndims,output.n,"MAIN",max_size);

nwind=(int*) malloc(sizeof(int)*output.ndims);
fwind=(int*) malloc(sizeof(int)*output.ndims);
jwind=(int*) malloc(sizeof(int)*output.ndims);

nloop=1;
if(copy==0){ /*data is split */
  jsect=(int*) malloc(sizeof(int)*output.ndims);
  nsect=(int*) malloc(sizeof(int)*output.ndims);
  fsect=(int*) malloc(sizeof(int)*output.ndims);
  jout=jsect; nout=nsect; fout=fsect;
  fin=fwind; jin=jwind; nin=nwind;
#ifndef SEP_MPI
  nloop=nassign;
#endif
}
else{ /*mirroring data*/
  jcopy=(int*) malloc(sizeof(int)*input.ndims);
  ncopy=(int*) malloc(sizeof(int)*input.ndims);
  fcopy=(int*) malloc(sizeof(int)*input.ndims);
  jout=jwind; nout=nwind; fout=fwind;
  fin=fcopy; jin=jcopy; nin=ncopy;
  for(i=0; i < input.ndims; i++) jin[i]=1;
  for(i=0; i < input.ndims; i++) nin[i]=1;
  for(i=0; i < input.ndims; i++) fin[i]=0;
  fin[axis-1]= assign[ impi-1]-1;
}

for(i=0; i < output.ndims; i++) jwind[i]=1;


while(0==do_sep_loop("MAIN",nwind,fwind)){

   for(isect=0;isect< nloop; isect++){

     nh=0;
     if(copy==1){
#ifdef SEP_MPI
       iassign=assign[sep_thread_num()-1]-1;
#else
       iassign=assign[isect]-1;
#endif
       if(0!=sep3dc_section_window(&input,iassign,nwind,fwind,
        jwind,nsect,fsect,jsect))
         seperr("trouble obtaining section window \n");
     } 
     else{
       for(i=0; i < output.ndims; i++){ nin[i]=nout[i]; fin[i]=fout[i];}
     }
     if(verb>0 && sep_thread_num()==0){
        fprintf(stderr,"WINDOW section=%d  ",assign[isect]);
        for(i=0; i < output.ndims ; i++)
          fprintf(stderr,"f[%d]=%d n[%d]=%d ",i+1,fwind[i],i+1,nwind[i]);
        fprintf(stderr,"\n");
      }

     if(0!= sep3dc_grab_headers("intag",&input,&nh,&nin[1],&fin[1],&jin[1]))
      seperr("trouble grabbing headers \n");
                                                                                   
     if(nh>0) buf_in=(char*) malloc(esize*nh*input.n[0]);

     if(0!=sep3dc_read_data("intag",&input,(void*)buf_in,nwind[0],
       fwind[0],jwind[0])) seperr("trouble reading data");
                                                                                   
                                                                                   
     if(copy==0){
       if(0!= sep3dc_collect_headers(&input,&output,&nhout))
         seperr("trouble collecting headers");
                                                                                   
                                                                                   
       if(sep_thread_num()==0) buf_out=(char*) malloc(esize*output.n[0]*nhout);
                                                                                   
       if(0!=sep3dc_collect_data(&input,&output,buf_in,buf_out))
         seperr("trouble collecting headers\n");
                                                                                   
       if(sep_thread_num()==0){
         if(0!= sep3dc_write_data("outtag",&output,(char*)buf_out,nwind,fwind,
            jwind,nh, write_head,write_grid))
         free(buf_out);
       }
    }
    else{
      if(isect==0){
        if(0!= sep3dc_pass_headers(&input, assign[0],0))
          seperr("trouble passing first header section\n");
        if(sep_thread_num()==0){
          if(0!= sep3dc_grab_nh(&input,&nhout))
            seperr("trouble grabbing nh");
          buf_out=(char*) malloc(esize*output.n[0]*nhout);
        }
        float_in=( float*) buf_in;
        if(sep_thread_num()==0) float_out=( float*) buf_out;
#ifdef SEP_MPI
     if(0!=sep3c_compress_data(&input,&output,buf_in,but_out))
        seperr("trouble compressing data \n");
#else
        for(i=0; i < nhout*output.n[0]/2; i++) buf_out[i]=buf_in[i];
#endif
      } 
      else{
#ifndef SEP_MPI
        for(i=0; i < nhout*output.n[0]/2; i++) buf_out[i]+=buf_in[i];
#endif
      }
      if(sep_thread_num()==0 && isect==nloop-1){
         if(0!= sep3dc_write_data("outtag",&output,(char*)buf_out,nwind,fwind,
            jwind,nh, write_head,write_grid))
         free(buf_out);
       }
    }
    if(nh==0) free(buf_in);
  }
}

#ifdef SEP_MPI
  MPI_Finalize();
#endif

  return(SUCCESS);
}
